@@ -632,7 +632,7 @@ lzw_read_byte (GifContext *context)
                         g_set_error (context->error,
                                      GDK_PIXBUF_ERROR,
                                      GDK_PIXBUF_ERROR_FAILED,
-                                     _("GIF image loader can't understand this image."));
+                                     _("GIF image loader cannot understand this image."));
                         return -2;
                         
 			if (ZeroDataBlock) {
@@ -770,13 +770,49 @@ gif_fill_in_lines (GifContext *context, guchar *dest, guchar v)
 	}
 }
 
-static void
-set_need_recomposite (gpointer data, gpointer user_data)
+/* Clips a rectancle to the base dimensions. Returns TRUE if the clipped rectangle is non-empty. */
+static gboolean
+clip_frame (GifContext *context, 
+            gint       *x, 
+            gint       *y, 
+            gint       *width, 
+            gint       *height)
 {
-        GdkPixbufFrame *frame = (GdkPixbufFrame *)data;
-        frame->need_recomposite = TRUE;
+        gint orig_x, orig_y;
+        
+        orig_x = *x;
+        orig_y = *y;
+	*x = MAX (0, *x);
+	*y = MAX (0, *y);
+	*width = MIN (context->width, orig_x + *width) - *x;
+	*height = MIN (context->height, orig_y + *height) - *y;
+
+	if (*width > 0 && *height > 0)
+		return TRUE;
+
+	/* The frame is completely off-bounds */
+
+	*x = 0;
+	*y = 0;
+	*width = 0;
+	*height = 0;
+
+        return FALSE;
 }
 
+/* Call update_func on the given rectangle, unless it is completely off-bounds */
+static void
+maybe_update (GifContext *context,
+              gint        x,
+              gint        y,
+              gint        width,
+              gint        height)
+{
+        if (clip_frame (context, &x, &y, &width, &height))
+                (*context->update_func) (context->frame->pixbuf, 
+                                         x, y, width, height,
+                                         context->user_data);
+}
 
 static int
 gif_get_lzw (GifContext *context)
@@ -863,25 +899,8 @@ gif_get_lzw (GifContext *context)
 
                 context->frame->bg_transparent = (context->gif89.transparent == context->background_index);
                 
-                {
-                        /* Update animation size */
-                        int w, h;
-                        
-                        context->animation->n_frames ++;
-                        context->animation->frames = g_list_append (context->animation->frames, context->frame);
-
-                        w = context->frame->x_offset +
-                                gdk_pixbuf_get_width (context->frame->pixbuf);
-                        h = context->frame->y_offset +
-                                gdk_pixbuf_get_height (context->frame->pixbuf);
-                        if (w  > context->animation->width || h > context->animation->height) {
-                                g_list_foreach (context->animation->frames, set_need_recomposite, NULL);
-                        }
-                        if (w > context->animation->width)
-                                context->animation->width = w;
-                        if (h > context->animation->height)
-                                context->animation->height = h;
-                }
+                context->animation->n_frames ++;
+                context->animation->frames = g_list_append (context->animation->frames, context->frame);
 
                 /* Only call prepare_func for the first frame */
 		if (context->animation->frames->next == NULL) { 
@@ -893,6 +912,7 @@ gif_get_lzw (GifContext *context)
                         /* Otherwise init frame with last frame */
                         GList *link;
                         GdkPixbufFrame *prev_frame;
+                        gint x, y, w, h;
                         
                         link = g_list_find (context->animation->frames, context->frame);
 
@@ -900,13 +920,43 @@ gif_get_lzw (GifContext *context)
 
                         gdk_pixbuf_gif_anim_frame_composite (context->animation, prev_frame);
 
-                        gdk_pixbuf_copy_area (prev_frame->composited,
-                                              context->frame->x_offset,
-                                              context->frame->y_offset,
-                                              gdk_pixbuf_get_width (context->frame->pixbuf),
-                                              gdk_pixbuf_get_height (context->frame->pixbuf),
-                                              context->frame->pixbuf,
-                                              0, 0);
+                        /* Composite failed */
+                        if (prev_frame->composited == NULL) {
+                                GdkPixbufFrame *frame = NULL;
+                                link = g_list_first (context->animation->frames);
+                                while (link != NULL) {
+                                        frame = (GdkPixbufFrame *)link->data;
+                                        if (frame != NULL) {
+                                                if (frame->pixbuf != NULL)
+                                                        g_object_unref (frame->pixbuf);
+                                                if (frame->composited != NULL)
+                                                        g_object_unref (frame->composited);
+                                                if (frame->revert != NULL)
+                                                        g_object_unref (frame->revert);
+                                                g_free (frame);
+                                        }
+                                        link = link->next;
+                                }
+                                
+                                g_list_free (context->animation->frames);
+                                context->animation->frames = NULL;
+                                
+                                g_set_error (context->error,
+                                             GDK_PIXBUF_ERROR,
+                                             GDK_PIXBUF_ERROR_INSUFFICIENT_MEMORY,
+                                             _("Not enough memory to composite a frame in GIF file"));
+                                return -2;
+                        }
+                    
+                        x = context->frame->x_offset;
+                        y = context->frame->y_offset;
+                        w = gdk_pixbuf_get_width (context->frame->pixbuf);
+                        h = gdk_pixbuf_get_height (context->frame->pixbuf);
+                        if (clip_frame (context, &x, &y, &w, &h))
+                                gdk_pixbuf_copy_area (prev_frame->composited,
+                                                      x, y, w, h,
+                                                      context->frame->pixbuf,
+                                                      0, 0);
                 }
         }
 
@@ -1005,37 +1055,29 @@ gif_get_lzw (GifContext *context)
         
 	if (bound_flag && context->update_func) {
 		if (lower_bound <= upper_bound && first_pass == context->draw_pass) {
-			(* context->update_func)
-				(context->frame->pixbuf,
-                                 context->frame->x_offset,
-                                 context->frame->y_offset + lower_bound,
-				 gdk_pixbuf_get_width (context->frame->pixbuf),
-				 upper_bound - lower_bound,
-				 context->user_data);
+                        maybe_update (context, 
+                                      context->frame->x_offset,
+                                      context->frame->y_offset + lower_bound,
+                                      gdk_pixbuf_get_width (context->frame->pixbuf),
+                                      upper_bound - lower_bound);
 		} else {
 			if (lower_bound <= upper_bound) {
-				(* context->update_func)
-					(context->frame->pixbuf,
-					 context->frame->x_offset,
-                                         context->frame->y_offset,
-					 gdk_pixbuf_get_width (context->frame->pixbuf),
-					 gdk_pixbuf_get_height (context->frame->pixbuf),
-					 context->user_data);
+                                maybe_update (context,
+                                              context->frame->x_offset,
+                                              context->frame->y_offset,
+                                              gdk_pixbuf_get_width (context->frame->pixbuf),
+                                              gdk_pixbuf_get_height (context->frame->pixbuf));
 			} else {
-				(* context->update_func)
-					(context->frame->pixbuf,
-					 context->frame->x_offset,
-                                         context->frame->y_offset,
-					 gdk_pixbuf_get_width (context->frame->pixbuf),
-					 upper_bound,
-					 context->user_data);
-				(* context->update_func)
-					(context->frame->pixbuf,
-					 context->frame->x_offset,
-                                         context->frame->y_offset + lower_bound,
-					 gdk_pixbuf_get_width (context->frame->pixbuf),
-					 gdk_pixbuf_get_height (context->frame->pixbuf) - lower_bound,
-					 context->user_data);
+                                maybe_update (context,
+                                              context->frame->x_offset,
+                                              context->frame->y_offset,
+                                              gdk_pixbuf_get_width (context->frame->pixbuf),
+                                              upper_bound);
+                                maybe_update (context,
+                                              context->frame->x_offset,
+                                              context->frame->y_offset + lower_bound,
+                                              gdk_pixbuf_get_width (context->frame->pixbuf),
+                                              gdk_pixbuf_get_height (context->frame->pixbuf) - lower_bound);
 			}
 		}
 	}
@@ -1164,7 +1206,10 @@ gif_init (GifContext *context)
         context->animation->bg_red = 0;
         context->animation->bg_green = 0;
         context->animation->bg_blue = 0;
-        
+
+        context->animation->width = context->width;
+        context->animation->height = context->height;
+
 	if (context->has_global_cmap) {
 		gif_set_get_colormap (context);
 	} else {
@@ -1200,23 +1245,10 @@ gif_get_frame_info (GifContext *context)
 	context->x_offset = LM_to_uint (buf[0], buf[1]);
 	context->y_offset = LM_to_uint (buf[2], buf[3]);
 
-	if (((context->frame_height + context->y_offset) > context->height) ||
-            ((context->frame_len + context->x_offset) > context->width)) {
-		/* All frames must fit in the image bounds */
-		context->state = GIF_DONE;
-
-                g_set_error (context->error,
-                             GDK_PIXBUF_ERROR,
-                             GDK_PIXBUF_ERROR_CORRUPT_IMAGE,
-                             _("GIF image contained a frame appearing outside the image bounds."));
-                
-		return -2;
-	}
-        
 	if (context->animation->frames == NULL &&
             context->gif89.disposal == 3) {
                 /* First frame can't have "revert to previous" as its
-                 * dispose mode. Silently use "retain" instead. 
+                 * dispose mode. Silently use "retain" instead.
                  */
                 context->gif89.disposal = 0;
 	}
@@ -1279,7 +1311,7 @@ gif_get_next_step (GifContext *context)
 		}
 
 		if (c == '!') {
-			/* Check the extention */
+			/* Check the extension */
 			gif_set_get_extension (context);
 			return 0;
 		}
@@ -1639,5 +1671,6 @@ MODULE_ENTRY (gif, fill_info) (GdkPixbufFormat *info)
 	info->description = N_("The GIF image format");
 	info->mime_types = mime_types;
 	info->extensions = extensions;
-	info->flags = 0;
+	info->flags = GDK_PIXBUF_FORMAT_THREADSAFE;
+	info->license = "LGPL";
 }
