@@ -27,6 +27,7 @@
 #ifndef __GDK_H__
 #define __GDK_H__
 
+#include <gdk/gdkcairo.h>
 #include <gdk/gdkcolor.h>
 #include <gdk/gdkcursor.h>
 #include <gdk/gdkdisplay.h>
@@ -53,9 +54,7 @@
 #include <gdk/gdkvisual.h>
 #include <gdk/gdkwindow.h>
 
-#ifdef __cplusplus
-extern "C" {
-#endif /* __cplusplus */
+G_BEGIN_DECLS
 
 
 /* Initialization, exit and events
@@ -67,6 +66,8 @@ void 	  gdk_init		   	(gint	   	*argc,
 					 gchar        ***argv);
 gboolean  gdk_init_check   	        (gint	   	*argc,
 					 gchar        ***argv);
+void gdk_add_option_entries_libgtk_only (GOptionGroup *group);
+void gdk_pre_parse_libgtk_only          (void);
 
 #ifndef GDK_DISABLE_DEPRECATED
 void  	  gdk_exit		   	(gint	    	 error_code);
@@ -115,6 +116,13 @@ GdkGrabStatus gdk_keyboard_grab      (GdkWindow    *window,
 				      gboolean      owner_events,
 				      guint32       time_);
 
+gboolean gdk_pointer_grab_info_libgtk_only (GdkDisplay *display,
+					    GdkWindow **grab_window,
+					    gboolean   *owner_events);
+gboolean gdk_keyboard_grab_info_libgtk_only (GdkDisplay *display,
+					     GdkWindow **grab_window,
+					     gboolean   *owner_events);
+
 #ifndef GDK_MULTIHEAD_SAFE
 void          gdk_pointer_ungrab     (guint32       time_);
 void          gdk_keyboard_ungrab    (guint32       time_);
@@ -144,7 +152,7 @@ void     gdk_rectangle_union     (GdkRectangle *src1,
 				  GdkRectangle *src2,
 				  GdkRectangle *dest);
 
-GType gdk_rectangle_get_type (void);
+GType gdk_rectangle_get_type (void) G_GNUC_CONST;
 
 #define GDK_TYPE_RECTANGLE (gdk_rectangle_get_type ())
 
@@ -199,9 +207,7 @@ void     gdk_threads_set_lock_functions   (GCallback enter_fn,
 #  define GDK_THREADS_LEAVE()
 #endif	/* !G_THREADS_ENABLED */
 
-#ifdef __cplusplus
-}
-#endif /* __cplusplus */
+G_END_DECLS
 
 
 #endif /* __GDK_H__ */
